//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Item/InstrumentItems.cpp
//! @brief     Implement class InstrumentItem and all its children
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Item/InstrumentItems.h"
#include "Base/Const/Units.h"
#include "Base/Pixel/RectangularPixel.h"
#include "Core/Simulation/DepthProbeSimulation.h"
#include "Device/Beam/Beam.h"
#include "Device/Coord/CoordSystem1D.h"
#include "Device/Coord/CoordSystem2D.h"
#include "Device/Detector/RectangularDetector.h"
#include "Device/Detector/SphericalDetector.h"
#include "GUI/Model/Item/BackgroundItems.h"
#include "GUI/Model/Item/BeamWavelengthItem.h"
#include "GUI/Model/Item/DetectorItems.h"
#include "GUI/Model/Item/ItemUtils.h"
#include "GUI/Model/Item/PointwiseAxisItem.h"
#include "GUI/Model/Item/RealDataItem.h"
#include "GUI/Model/Item/RectangularDetectorItem.h"
#include "GUI/Model/Item/SphericalDetectorItem.h"
#include "GUI/Model/MakeItem/BackgroundItemCatalog.h"
#include "GUI/Model/MakeItem/DetectorItemCatalog.h"
#include "GUI/Model/MakeItem/InstrumentItemCatalog.h"
#include "GUI/Model/XML/Serializer.h"
#include "GUI/Util/CoordName.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/String.h"

namespace {

BasicAxisItem* addAxisGroupProperty(SessionItem* parent, const QString& tag)
{
    auto* axisItem = parent->addProperty<BasicAxisItem>(tag);
    axisItem->setToolTip("Incoming alpha range [deg]");
    axisItem->binsItem()->setToolTip("Number of points in scan");
    axisItem->lowerBoundItem()->setToolTip("Starting value [deg]");
    axisItem->upperBoundItem()->setToolTip("Ending value [deg]");

    axisItem->setTitle("alpha_i");
    axisItem->setLowerBound(0.0);
    axisItem->setUpperBound(10.0);

    return axisItem;
}

void setBeamDistribution(ParameterDistribution::WhichParameter which,
                         const BeamDistributionItem& item, ISimulation& simulation)
{
    auto P_par_distr = item.getParameterDistributionForName(which);
    if (P_par_distr)
        simulation.addParameterDistribution(*P_par_distr);
}

} // namespace

//  ************************************************************************************************
//  class InstrumentItem
//  ************************************************************************************************

InstrumentItem::InstrumentItem()
    : m_withPolarizerAnalyzer(false)
{
    m_id = QUuid::createUuid().toString();
    m_analyzerEfficiency.init("Efficiency", "", 0.0, Unit::unitless, 3, RealLimits::limitless(),
                              "efficiency");
    m_analyzerTotalTransmission.init("Transmission", "", 1.0, Unit::unitless, "transmission");
    m_polarization.init("Polarization", "", Unit::unitless, "polarization");
    m_analyzerDirection.init("Analyzer direction", "", Unit::unitless, "analyzerDirection");
    m_background.init<BackgroundItemCatalog>("Background", "", "background");
}

InstrumentItem* InstrumentItem::createCopy() const
{
    const auto type = InstrumentItemCatalog::type(this);
    auto* copy = InstrumentItemCatalog::create(type);
    GUI::Util::copyContents(this, copy);
    return copy;
}

QString InstrumentItem::id() const
{
    return m_id;
}

void InstrumentItem::setId(const QString& id)
{
    m_id = id;
}

void InstrumentItem::setInstrumentName(const QString& instrumentName)
{
    m_name = instrumentName;
}

QString InstrumentItem::instrumentName() const
{
    return m_name;
}

QString InstrumentItem::description() const
{
    return m_description;
}

void InstrumentItem::setDescription(const QString& description)
{
    m_description = description;
}

BeamItem* InstrumentItem::beamItem() const
{
    return m_beam.get();
}

BackgroundItem* InstrumentItem::backgroundItem() const
{
    return m_background.get();
}

SelectionDescriptor<BackgroundItem*> InstrumentItem::backgroundSelection() const
{
    return m_background;
}

bool InstrumentItem::alignedWith(const RealDataItem* item) const
{
    return shape() == item->shape();
}

bool InstrumentItem::withPolarizerAnalyzer() const
{
    return m_withPolarizerAnalyzer;
}

void InstrumentItem::setWithPolarizerAnalyzer(bool with)
{
    m_withPolarizerAnalyzer = with;
}

template <typename T> void InstrumentItem::addBeam()
{
    m_beam.reset(new T());
}

template <typename T> T* InstrumentItem::beam() const
{
    return dynamic_cast<T*>(m_beam.get());
}

void InstrumentItem::serialize(Serializer& s)
{
    s.assertVersion(0);
    s.rw("id", m_id);
    s.rw("name", m_name);
    s.rw("description", m_description);
    s.rw("pa", m_withPolarizerAnalyzer);
    s.rw(m_polarization);
    s.rw(m_analyzerDirection);
    s.rw(m_analyzerEfficiency);
    s.rw(m_analyzerTotalTransmission);
    s.rw<BackgroundItemCatalog>(m_background);
    s.rw("beam", *m_beam);
}

//  ************************************************************************************************
//  class SpecularInstrumentItem
//  ************************************************************************************************

SpecularInstrumentItem::SpecularInstrumentItem()
{
    addBeam<SpecularBeamItem>();
}

SpecularBeamItem* SpecularInstrumentItem::beamItem() const
{
    return beam<SpecularBeamItem>();
}

std::vector<int> SpecularInstrumentItem::shape() const
{
    auto* const axis_item = beamItem()->inclinationAxis();
    return {axis_item->binCount()};
}

void SpecularInstrumentItem::updateToRealData(const RealDataItem* item)
{
    if (shape().size() != item->shape().size())
        throw Error("Error in SpecularInstrumentItem::updateToRealData: The type "
                    "of instrument is incompatible with passed data shape.");

    const auto& data = item->nativeOutputData()->axis(0);
    beamItem()->updateToData(data, item->nativeDataUnits());
}

bool SpecularInstrumentItem::alignedWith(const RealDataItem* item) const
{
    const QString native_units = item->nativeDataUnits();
    if (native_units == "nbins") {
        return beamItem()->inclinationAxis()->modelType() == BasicAxisItem::M_TYPE
               && shape() == item->shape();
    }
    auto* axis_item = dynamic_cast<PointwiseAxisItem*>(beamItem()->inclinationAxis());
    if (!axis_item)
        return false;
    if (axis_item->getUnitsLabel() != native_units)
        return false;

    const auto* instrument_axis = axis_item->axis();
    if (!instrument_axis)
        return false;

    if (!item->hasNativeData())
        return false;

    const auto& native_axis = item->nativeOutputData()->axis(0);
    return *instrument_axis == native_axis;
}

ICoordSystem* SpecularInstrumentItem::createCoordSystem() const
{
    std::unique_ptr<Instrument> instrument(new Instrument);
    instrument->setBeam(*beamItem()->createBeam());
    // \todo setPolFilters ?
    auto* axis_item = beamItem()->inclinationAxis();
    if (auto* pointwise_axis = dynamic_cast<PointwiseAxisItem*>(axis_item)) {
        if (!pointwise_axis->axis()) // workaround for loading project
            return nullptr;
        Axes::Coords native_units =
            GUI::Util::CoordName::coordFromName(pointwise_axis->getUnitsLabel());
        return new AngularReflectometryCoordinates(instrument->beam().wavelength(),
                                                   *pointwise_axis->axis(), native_units);
    }

    return new AngularReflectometryCoordinates(instrument->beam().wavelength(),
                                               *axis_item->createAxis(1.0), Axes::Coords::DEGREES);
}

QString SpecularInstrumentItem::instrumentType() const
{
    return "Specular";
}

QString SpecularInstrumentItem::defaultName() const
{
    return "Specular";
}

//  ************************************************************************************************
//  class DepthProbeInstrumentItem
//  ************************************************************************************************

DepthProbeInstrumentItem::DepthProbeInstrumentItem()
{
    addBeam<SpecularBeamItem>();

    auto* axisItem = beamItem()->inclinationAxis();
    axisItem->setLowerBound(0.0);
    axisItem->setUpperBound(1.0);
    axisItem->setBinCount(500);

    m_zAxis.initNbins("Nbins", "Number of points in scan across sample bulk");
    m_zAxis.initMin("Min", "Starting value below sample horizon", -100.0, Unit::nanometer);
    m_zAxis.initMax("Max", "Ending value above sample horizon", 100.0, Unit::nanometer);
}

void DepthProbeInstrumentItem::serialize(Serializer& s)
{
    s.assertVersion(0);
    s.rwBaseClass<InstrumentItem>("InstrumentItem", this);
    s.rw("zAxis", m_zAxis);
}

SpecularBeamItem* DepthProbeInstrumentItem::beamItem() const
{
    return beam<SpecularBeamItem>();
}

std::vector<int> DepthProbeInstrumentItem::shape() const
{
    return std::vector<int>(); // no certain shape to avoid linking to real data
}

void DepthProbeInstrumentItem::updateToRealData(const RealDataItem*)
{
    throw std::runtime_error("DepthProbeInstrumentItem::updateToRealData()");
}

QString DepthProbeInstrumentItem::defaultName() const
{
    return "DepthProbe";
}

QString DepthProbeInstrumentItem::instrumentType() const
{
    return "Depth probe";
}

std::unique_ptr<DepthProbeSimulation> DepthProbeInstrumentItem::createSimulation() const
{
    std::unique_ptr<DepthProbeSimulation> simulation = std::make_unique<DepthProbeSimulation>();

    auto* const axis_item = beamItem()->inclinationAxis();
    auto axis = axis_item->createAxis(Units::deg);

    simulation->setBeamParameters(beamItem()->wavelength(), static_cast<int>(axis->size()),
                                  axis->lowerBound(), axis->upperBound());

    simulation->setZSpan(m_zAxis.nbins(), m_zAxis.min(), m_zAxis.max());

    setBeamDistribution(ParameterDistribution::BeamWavelength, *beamItem()->wavelengthItem(),
                        *simulation);

    setBeamDistribution(ParameterDistribution::BeamInclinationAngle,
                        *beamItem()->inclinationAngleItem(), *simulation);

    return simulation;
}

ICoordSystem* DepthProbeInstrumentItem::createCoordSystem() const
{
    // \todo setPolFilters ?
    return createSimulation()->createCoordSystem();
}

//  ************************************************************************************************
//  class Instrument2DItem
//  ************************************************************************************************

Instrument2DItem::Instrument2DItem()
{
    addBeam<GISASBeamItem>();
    m_detector.init<DetectorItemCatalog>("Detector", "", "detector");
}

void Instrument2DItem::serialize(Serializer& s)
{
    s.assertVersion(0);
    s.rwBaseClass<InstrumentItem>("InstrumentItem", this);
    s.rw<DetectorItemCatalog>(m_detector);
}

DetectorItem* Instrument2DItem::detector() const
{
    return m_detector.get();
}

SelectionDescriptor<DetectorItem*> Instrument2DItem::detectorSelection() const
{
    return m_detector;
}

void Instrument2DItem::importMasks(const MaskContainerItem* maskContainer)
{
    detector()->importMasks(maskContainer);
}

std::unique_ptr<Instrument> Instrument2DItem::createInstrument() const
{
    std::unique_ptr<Instrument> instrument(new Instrument);
    instrument->setBeam(*beamItem()->createBeam());
    instrument->setDetector(*detector()->createDetector());

    // TODO: Why not check "if (withPolarizerAnalyzer())"? But if it is inserted, some of the unit
    // tests fail (because in the unit tests this flag is never set to true)
    // The Unit test are e.g. TestGuiStd/Std/MagneticCylinders
    instrument->setPolFilters(m_polarization, m_analyzerDirection, m_analyzerEfficiency,
                              m_analyzerTotalTransmission);

    return instrument;
}

//  ************************************************************************************************
//  class GISASInstrumentItem
//  ************************************************************************************************

std::vector<int> GISASInstrumentItem::shape() const
{
    auto* detector_item = detector();
    return {detector_item->xSize(), detector_item->ySize()};
}

void GISASInstrumentItem::updateToRealData(const RealDataItem* item)
{
    if (!item)
        return;

    const auto data_shape = item->shape();
    if (shape().size() != data_shape.size())
        throw Error("Error in GISASInstrumentItem::updateToRealData: The type of "
                    "instrument is incompatible with passed data shape.");
    detector()->setXSize(data_shape[0]);
    detector()->setYSize(data_shape[1]);
}

QString GISASInstrumentItem::defaultName() const
{
    return "GISAS";
}

QString GISASInstrumentItem::instrumentType() const
{
    return "GISAS";
}

ICoordSystem* GISASInstrumentItem::createCoordSystem() const
{
    const auto instrument = createInstrument();
    instrument->initDetector();
    return instrument->createScatteringCoords();
}

//  ************************************************************************************************
//  class OffSpecularInstrumentItem
//  ************************************************************************************************

OffSpecularInstrumentItem::OffSpecularInstrumentItem()
{
    m_alphaAxis.initNbins("Nbins", "Number of points in scan across sample bulk");
    m_alphaAxis.initMin("Min", "Starting value below sample horizon", -100.0, Unit::nanometer);
    m_alphaAxis.initMax("Max", "Ending value above sample horizon", 100.0, Unit::nanometer);

    auto* beam_item = dynamic_cast<GISASBeamItem*>(beamItem());
    beam_item->setInclinationAngleGetter([=]() { return m_alphaAxis.min(); });
}

void OffSpecularInstrumentItem::serialize(Serializer& s)
{
    s.assertVersion(0);
    s.rwBaseClass<Instrument2DItem>("Instrument2DItem", this);
    s.rw("alphaAxis", m_alphaAxis);
}

std::vector<int> OffSpecularInstrumentItem::shape() const
{
    return {(int)m_alphaAxis.nbins(), detector()->ySize()};
}

void OffSpecularInstrumentItem::updateToRealData(const RealDataItem* dataItem)
{
    if (!dataItem)
        return;

    const auto data_shape = dataItem->shape();
    if (shape().size() != data_shape.size())
        throw Error("Error in OffSpecularInstrumentItem::updateToRealData: The type of "
                    "instrument is incompatible with passed data shape.");

    m_alphaAxis.setNbins(data_shape[0]);
    detector()->setYSize(data_shape[1]);
}

QString OffSpecularInstrumentItem::defaultName() const
{
    return "OffSpecular";
}

QString OffSpecularInstrumentItem::instrumentType() const
{
    return "Off specular";
}

ICoordSystem* OffSpecularInstrumentItem::createCoordSystem() const
{
    const auto instrument = createInstrument();
    instrument->initDetector();
    const auto* const detector2d = dynamic_cast<const IDetector2D*>(instrument->getDetector());
    const auto axes = detector2d->axesClippedToRegionOfInterest();
    ASSERT(axes.size() == 2);
    const IAxis& yAxis = *axes[1];
    const Beam& beam = instrument->beam();
    const auto alphaAxis = std::make_unique<FixedBinAxis>(std::string(), m_alphaAxis.nbins(),
                                                          m_alphaAxis.min() * Units::deg,
                                                          m_alphaAxis.max() * Units::deg);

    if (const auto* rectDetector = dynamic_cast<const RectangularDetector*>(detector2d)) {
        std::unique_ptr<RectangularPixel> detectorPixel(rectDetector->regionOfInterestPixel());
        return OffSpecularCoordinates::createForRectangularDetector(beam.direction(), *alphaAxis,
                                                                    *detectorPixel, yAxis);
    }
    if (dynamic_cast<const SphericalDetector*>(detector2d))
        return OffSpecularCoordinates::createForSphericalDetector(beam.direction(), *alphaAxis,
                                                                  yAxis);

    ASSERT(0);
}

//  ************************************************************************************************
//  class InstrumentItems
//  ************************************************************************************************

InstrumentItems::~InstrumentItems()
{
    qDeleteAll(m_instruments);
}

void InstrumentItems::clear()
{
    qDeleteAll(m_instruments);
    m_instruments.clear();
}

void InstrumentItems::serialize(Serializer& s)
{
    s.assertVersion(0);
    s.rw<InstrumentItemCatalog>("InstrumentItems", m_instruments);
}

InstrumentItem* InstrumentItems::insertCopy(const InstrumentItem& source)
{
    auto* copy = source.createCopy();
    copy->setId(QUuid::createUuid().toString());
    m_instruments << copy;
    return copy;
}

QVector<InstrumentItem*> InstrumentItems::instrumentItems() const
{
    return m_instruments;
}

QStringList InstrumentItems::instrumentNames() const
{
    QStringList existingNames;
    for (const auto* item : instrumentItems())
        existingNames << item->instrumentName();
    return existingNames;
}

QString InstrumentItems::suggestInstrumentName(const QString& baseName) const
{
    return GUI::Util::String::suggestName(instrumentNames(), baseName);
}

QVector<InstrumentItem*>
InstrumentItems::instrumentItems(const std::function<bool(const InstrumentItem*)>& accept) const
{
    QVector<InstrumentItem*> result;
    for (auto* p : instrumentItems())
        if (accept(p))
            result << p;

    return result;
}

QVector<Instrument2DItem*> InstrumentItems::instrument2DItems() const
{
    QVector<Instrument2DItem*> result;
    for (auto* p : instrumentItems())
        if (auto* p2D = dynamic_cast<Instrument2DItem*>(p))
            result << p2D;

    return result;
}

InstrumentItem* InstrumentItems::findInstrumentById(const QString& instrumentId) const
{
    for (auto* instrument : instrumentItems())
        if (instrument->id() == instrumentId)
            return instrument;

    return nullptr;
}

bool InstrumentItems::instrumentExists(const QString& instrumentId) const
{
    return findInstrumentById(instrumentId) != nullptr;
}

void InstrumentItems::removeInstrument(InstrumentItem* instrument)
{
    m_instruments.removeAll(instrument);
    delete instrument;
}
