//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Item/SphericalDetectorItem.cpp
//! @brief     Implements class SphericalDetectorItem
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Item/SphericalDetectorItem.h"
#include "Base/Const/Units.h"
#include "Device/Detector/SphericalDetector.h"
#include "GUI/Model/MakeItem/ResolutionFunctionItemCatalog.h"
#include "GUI/Model/XML/Serializer.h"
#include "ResolutionFunctionItems.h"

namespace {
void initResolutionFunction(ResolutionFunctionItem* newFunc, const ResolutionFunctionItem*)
{
    newFunc->setUnit(Unit::degree);
}
} // namespace

SphericalDetectorItem::SphericalDetectorItem()
{
    m_resolutionFunction.init<ResolutionFunctionItemCatalog>(
        "Resolution function", "Detector resolution function", "resolutionFunction",
        initResolutionFunction);

    m_phiAxis.initNbins("Nbins", "Number of phi-axis bins");
    m_phiAxis.initMin("Min", "Lower edge of first phi-bin", -1.0, Unit::degree);
    m_phiAxis.initMax("Max", "Upper edge of last phi-bin", 1.0, Unit::degree);

    m_alphaAxis.initNbins("Nbins", "Number of alpha-axis bins");
    m_alphaAxis.initMin("Min", "Lower edge of first alpha-bin", 0.0, Unit::degree);
    m_alphaAxis.initMax("Max", "Upper edge of last alpha-bin", 2.0, Unit::degree);
}

void SphericalDetectorItem::serialize(Serializer& s)
{
    s.assertVersion(0);
    // base class members
    s.rw("masks", m_maskItems);
    s.rw<ResolutionFunctionItemCatalog>(m_resolutionFunction);

    // own members
    s.rw("phiAxis", m_phiAxis);
    s.rw("alphaAxis", m_alphaAxis);
}

std::unique_ptr<IDetector2D> SphericalDetectorItem::createDomainDetector() const
{
    const int n_x = m_phiAxis.nbins();
    const double x_min = Units::deg2rad(m_phiAxis.min());
    const double x_max = Units::deg2rad(m_phiAxis.max());

    const int n_y = m_alphaAxis.nbins();
    const double y_min = Units::deg2rad(m_alphaAxis.min());
    const double y_max = Units::deg2rad(m_alphaAxis.max());

    return std::make_unique<SphericalDetector>(n_x, x_min, x_max, n_y, y_min, y_max);
}

int SphericalDetectorItem::xSize() const
{
    return m_phiAxis.nbins();
}

int SphericalDetectorItem::ySize() const
{
    return m_alphaAxis.nbins();
}

void SphericalDetectorItem::setXSize(size_t nx)
{
    m_phiAxis.setNbins(nx);
}

void SphericalDetectorItem::setYSize(size_t ny)
{
    m_alphaAxis.setNbins(ny);
}

double SphericalDetectorItem::axesToDomainUnitsFactor() const
{
    return Units::deg;
}
